/* SPDX-License-Identifier: MIT OR X11
 *
 * Copyright © 2024 Enrico Weigelt, metux IT consult <info@metux.net>
 */
#ifndef _XSERVER_CALLBACK_PRIV_H
#define _XSERVER_CALLBACK_PRIV_H

#include "callback.h"

typedef struct _CallbackList *CallbackListPtr;

void InitCallbackManager(void);
void DeleteCallbackManager(void);

#endif /* _XSERVER_CALLBACK_PRIV_H */
